@@ -14,6 +14,9 @@ from shutil import copyfileobj
 from zlib import decompress
 from zlib import error as ZlibError
 
+from PIL import Image, ImageCms
+from PIL.TiffTags import TAGS_V2 as TIFF_TAGS
+
 from pikepdf import (
     Array,
     Dictionary,
@@ -22,6 +25,7 @@ from pikepdf import (
     PdfError,
     Stream,
     StreamDecodeLevel,
+    String,
     jbig2,
 )
 
@@ -39,26 +43,24 @@ class NotExtractableError(Exception):
 
 
 def array_str(value):
-    if isinstance(value, (list, Array)):
-        return [str(item) for item in value]
-    if isinstance(value, Name):
-        return [str(value)]
-    raise NotImplementedError(value)
+    """Simplify pikepdf objects to array of str. Keep Streams intact."""
 
+    def _array_str(item):
+        if isinstance(item, (list, Array)):
+            return [_array_str(subitem) for subitem in item]
+        elif isinstance(item, (Stream, bytes, int)):
+            return item
+        elif isinstance(item, (Name, str)):
+            return str(item)
+        elif isinstance(item, (String)):
+            return bytes(item)
+        else:
+            raise NotImplementedError(value)
 
-def array_str_colorspace(value):
-    if isinstance(value, (list, Array)):
-        items = [item for item in value]
-        if len(items) == 4 and items[0] == '/Indexed':
-            result = [str(items[n]) for n in range(3)]
-            result.append(bytes(items[3]))
-            return result
-        if len(items) == 2 and items[0] == '/ICCBased':
-            result = [str(items[0]), items[1]]
-            return result
-        return array_str(items)
-
-    return array_str(value)
+    result = _array_str(value)
+    if not isinstance(result, list):
+        result = [result]
+    return result
 
 
 def dict_or_array_dict(value):
@@ -83,7 +85,8 @@ def metadata_from_obj(obj, name, type_, default):
 
 class PdfImageBase(ABC):
 
-    SIMPLE_COLORSPACES = ('/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray')
+    SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
+    MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
 
     @abstractmethod
     def _metadata(self, name, type_, default):
@@ -112,7 +115,7 @@ class PdfImageBase(ABC):
     @property
     def _colorspaces(self):
         """Colorspace (low-level)"""
-        return self._metadata('ColorSpace', array_str_colorspace, [])
+        return self._metadata('ColorSpace', array_str, [])
 
     @property
     def filters(self):
@@ -130,15 +133,14 @@ class PdfImageBase(ABC):
         if self.image_mask:
             return None  # Undefined for image masks
         if self._colorspaces:
-            if self._colorspaces[0] in self.SIMPLE_COLORSPACES:
+            if self._colorspaces[0] in self.MAIN_COLORSPACES:
                 return self._colorspaces[0]
-            if self._colorspaces[0] in ('/DeviceCMYK', '/ICCBased'):
-                return self._colorspaces[0]
-            if (
-                self._colorspaces[0] == '/Indexed'
-                and self._colorspaces[1] in self.SIMPLE_COLORSPACES
-            ):
-                return self._colorspaces[1]
+            if self._colorspaces[0] == '/Indexed':
+                subspace = self._colorspaces[1]
+                if isinstance(subspace, str) and subspace in self.MAIN_COLORSPACES:
+                    return subspace
+                if isinstance(subspace, list) and subspace[0] == '/ICCBased':
+                    return subspace[0]
         raise NotImplementedError(
             "not sure how to get colorspace: " + repr(self._colorspaces)
         )
@@ -234,15 +236,28 @@ class PdfImageBase(ABC):
             _idx, base, _hival, lookup = self._colorspaces
         except ValueError as e:
             raise ValueError('Not sure how to interpret this palette') from e
-        base = str(base)
-        _hival = int(_hival)
+        iccobj = None
+        if self.icc:
+            iccobj = base[1]
+            base = str(base[0])
+        else:
+            base = str(base)
         lookup = bytes(lookup)
-        if not base in self.SIMPLE_COLORSPACES:
+        if not base in self.MAIN_COLORSPACES:
             raise NotImplementedError("not sure how to interpret this palette")
         if base == '/DeviceRGB':
             base = 'RGB'
         elif base == '/DeviceGray':
             base = 'L'
+        elif base == '/DeviceCMYK':
+            base = 'CMYK'
+        elif base == '/ICCBased':
+            if iccobj['/N'] == 3:
+                base = 'RGB'
+            elif iccobj['/N'] == 4:
+                base = 'CMYK'
+            elif iccobj['/N'] == 1:
+                base = 'L'
         return base, lookup
 
     @abstractmethod
@@ -305,6 +320,9 @@ class PdfImage(PdfImageBase):
         self.obj = obj
         self._icc = None
 
+    def __eq__(self, other):
+        return self.obj == other.obj
+
     @classmethod
     def _from_pil_image(cls, *, pdf, page, name, image):  # pragma: no cover
         """Insert a PIL image into a PDF (rudimentary)
@@ -341,6 +359,15 @@ class PdfImage(PdfImageBase):
         """``False`` for image XObject"""
         return False
 
+    @property
+    def _iccstream(self):
+        if self.colorspace == '/ICCBased':
+            if not self.indexed:
+                return self._colorspaces[1]
+            assert isinstance(self._colorspaces[1], list)
+            return self._colorspaces[1][1]
+        raise NotImplementedError("Don't know how to find ICC stream for image")
+
     @property
     def icc(self):
         """If an ICC profile is attached, return a Pillow object that describe it.
@@ -350,12 +377,10 @@ class PdfImage(PdfImageBase):
         Returns:
             PIL.ImageCms.ImageCmsProfile
         """
-        from PIL import ImageCms
-
-        if self.colorspace != '/ICCBased':
+        if self.colorspace not in ('/ICCBased', '/Indexed'):
             return None
         if not self._icc:
-            iccstream = self._colorspaces[1]
+            iccstream = self._iccstream
             iccbuffer = iccstream.get_stream_buffer()
             iccbytesio = BytesIO(iccbuffer)
             self._icc = ImageCms.ImageCmsProfile(iccbytesio)
@@ -395,8 +420,10 @@ class PdfImage(PdfImageBase):
 
         if filters == ['/CCITTFaxDecode']:
             if self.colorspace == '/ICCBased':
-                raise UnsupportedImageTypeError("Cannot direct-extract CCITT + ICC")
-            stream.write(self._generate_ccitt_header(data))
+                icc = self._iccstream.read_bytes()
+            else:
+                icc = None
+            stream.write(self._generate_ccitt_header(data, icc=icc))
             stream.write(data)
             return '.tif'
         elif filters == ['/DCTDecode'] and (
@@ -408,8 +435,6 @@ class PdfImage(PdfImageBase):
         raise NotExtractableError()
 
     def _extract_transcoded(self):
-        from PIL import Image
-
         im = None
         if self.mode == 'RGB' and self.bits_per_component == 8:
             # No point in accessing the buffer here, size qpdf decodes to 3-byte
@@ -461,18 +486,13 @@ class PdfImage(PdfImageBase):
         return im
 
     def _extract_to_stream(self, *, stream):
-        """Attempt to extract the image directly to a usable image file
+        """Attempt to extract the image to a stream
 
         If possible, the compressed data is extracted and inserted into
         a compressed image file format without transcoding the compressed
         content. If this is not possible, the data will be decompressed
         and extracted to an appropriate format.
 
-        Because it is not known until attempted what image format will be
-        extracted, users should not assume what format they are getting back.
-        When saving the image to a file, use a temporary filename, and then
-        rename the file to its final name based on the returned file extension.
-
         Args:
             stream: Writable stream to write data to
 
@@ -559,8 +579,6 @@ class PdfImage(PdfImageBase):
         Returns:
             PIL.Image.Image
         """
-        from PIL import Image
-
         try:
             bio = BytesIO()
             self._extract_direct(stream=bio)
@@ -575,7 +593,7 @@ class PdfImage(PdfImageBase):
 
         return im
 
-    def _generate_ccitt_header(self, data):
+    def _generate_ccitt_header(self, data, icc=None):
         """Construct a CCITT G3 or G4 header from the PDF metadata"""
         # https://stackoverflow.com/questions/2641770/
         # https://www.itu.int/itudoc/itu-t/com16/tiff-fx/docs/tiff6.pdf
@@ -603,25 +621,59 @@ class PdfImage(PdfImageBase):
         photometry = 1 if black_is_one else 0
 
         img_size = len(data)
-        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H' + 'HHLL' * 8 + 'L'
-        # fmt: off
+        tiff_header_struct = '<' + '2s' + 'H' + 'L' + 'H'
+
+        tag_keys = {tag.name: key for key, tag in TIFF_TAGS.items()}
+        ifd_struct = '<HHLL'
+
+        if icc is None:
+            icc = b''
+
+        ifds = []
+
+        def header_length(ifd_count):
+            return (
+                struct.calcsize(tiff_header_struct)
+                + struct.calcsize(ifd_struct) * ifd_count
+                + 4
+            )
+
+        def add_ifd(tag_name, data, count=1):
+            key = tag_keys[tag_name]
+            typecode = TIFF_TAGS[key].type
+            ifds.append((key, typecode, count, data))
+
+        image_offset = None
+        add_ifd('ImageWidth', self.width)
+        add_ifd('ImageLength', self.height)
+        add_ifd('BitsPerSample', 1)
+        add_ifd('Compression', ccitt_group)
+        add_ifd('PhotometricInterpretation', int(photometry))
+        add_ifd('StripOffsets', lambda: image_offset)
+        add_ifd('RowsPerStrip', self.height)
+        add_ifd('StripByteCounts', img_size)
+
+        icc_offset = 0
+        if icc:
+            add_ifd('ICCProfile', lambda: icc_offset, count=len(icc))
+
+        icc_offset = header_length(len(ifds))
+        image_offset = icc_offset + len(icc)
+
+        ifd_args = [(arg() if callable(arg) else arg) for ifd in ifds for arg in ifd]
         tiff_header = struct.pack(
-            tiff_header_struct,
+            (tiff_header_struct + ifd_struct[1:] * len(ifds) + 'L'),
             b'II',  # Byte order indication: Little endian
             42,  # Version number (always 42)
             8,  # Offset to first IFD
-            8,  # Number of tags in IFD
-            256, 4, 1, self.width,  # ImageWidth, LONG, 1, width
-            257, 4, 1, self.height,  # ImageLength, LONG, 1, length
-            258, 3, 1, 1,  # BitsPerSample, SHORT, 1, 1
-            259, 3, 1, ccitt_group,  # Compression, SHORT, 1, 4 = CCITT Group 4 fax encoding
-            262, 3, 1, int(photometry),  # Thresholding, SHORT, 1, 0 = WhiteIsZero
-            273, 4, 1, struct.calcsize(tiff_header_struct),  # StripOffsets, LONG, 1, length of header
-            278, 4, 1, self.height,
-            279, 4, 1, img_size,  # StripByteCounts, LONG, 1, size of image
-            0  # last IFD
+            len(ifds),  # Number of tags in IFD
+            *ifd_args,
+            0,  # Last IFD
         )
-        # fmt: on
+
+        if icc:
+            tiff_header += icc
+
         return tiff_header
 
     def show(self):
@@ -644,7 +696,11 @@ class PdfImage(PdfImageBase):
 class PdfJpxImage(PdfImage):
     def __init__(self, obj):
         super().__init__(obj)
-        self.pil = self.as_pil_image()
+        self._jpxpil = self.as_pil_image()
+
+    def __eq__(self, other):  # pylint: disable=useless-super-delegation
+        # self._jpxpil is not relevant to equality
+        return super().__eq__(other)
 
     def _extract_direct(self, *, stream):
         data, filters = self._unstack_compression(
@@ -662,9 +718,9 @@ class PdfJpxImage(PdfImage):
         super_colorspaces = super()._colorspaces
         if super_colorspaces:
             return super_colorspaces
-        if self.pil.mode == 'L':
+        if self._jpxpil.mode == 'L':
             return ['/DeviceGray']
-        elif self.pil.mode == 'RGB':
+        elif self._jpxpil.mode == 'RGB':
             return ['/DeviceRGB']
         raise NotImplementedError('Complex JP2 colorspace')
 
@@ -736,6 +792,14 @@ class PdfInlineImage(PdfImageBase):
         self.obj = reparsed_obj
         self.pil = None
 
+    def __eq__(self, other):
+        if self.obj == other.obj and (
+            self._data._inline_image_raw_bytes()
+            == other._data._inline_image_raw_bytes()
+        ):
+            return True
+        return False
+
     @classmethod
     def _unparse_obj(cls, obj):
         if isinstance(obj, Object):
@@ -757,17 +821,17 @@ class PdfInlineImage(PdfImageBase):
 
     def unparse(self):
         tokens = []
-        tokens.append(b'BI')
+        tokens.append(b'BI\n')
         metadata = []
         for metadata_obj in self._image_object:
             unparsed = self._unparse_obj(metadata_obj)
             assert isinstance(unparsed, bytes)
             metadata.append(unparsed)
         tokens.append(b' '.join(metadata))
-        tokens.append(b'ID')
+        tokens.append(b'\nID\n')
         tokens.append(self._data._inline_image_raw_bytes())
         tokens.append(b'EI')
-        return b'\n'.join(tokens)
+        return b''.join(tokens)
 
     @property
     def is_inline(self):
@@ -775,7 +839,7 @@ class PdfInlineImage(PdfImageBase):
 
     @property
     def icc(self):
-        raise ValueError("Inline images may not have ICC profiles")
+        raise NotImplementedError("Inline images with ICC profiles are not supported")
 
     def __repr__(self):
         mode = '?'
